import asyncio
import base64
import multiprocessing
import os
import threading
from io import BytesIO
from pathlib import Path

import numpy as np
import tifffile
from PIL import Image
from flet_extended_interactive_viewer import FletExtendedInteractiveViewer

from cellsepi.backend.main_window.data_util import convert_tiffs_to_png_parallel
from cellsepi.backend.main_window.expert_mode.listener import ProgressEvent, OnPipelineChangeEvent
from cellsepi.backend.main_window.expert_mode.module import *
from cellsepi.backend.main_window.image_tuning import auto_adjust
from cellsepi.frontend.drawing_window.gui_drawing import open_qt_window


class Review(Module, ABC):
    mask_color = (255, 0, 0)
    mask_opacity = 128
    outline_color = (0, 255, 0)
    _instances = []
    _gui_config = ModuleGuiConfig("Review",Categories.MANUAL,"This module allows you to manually review and edit masks. Also you can create new masks when no mask are given.")
    def __init__(self, module_id: str) -> None:
        #regular modul
        super().__init__(module_id)
        self.inputs = {
            "image_paths": InputPort("image_paths", dict),
            "mask_paths": InputPort("mask_paths", dict,True),
        }
        self.outputs = {
            "mask_paths": OutputPort("mask_paths", dict),
        }
        self._on_settings_dismiss = self.dismiss
        self.user_segmentation_channel: str = "2"
        self.user_2_5d = False
        self.user_mask_suffix = "_seg"
        #for the own settings stack
        self._icon_x = {}
        self._icon_check = {}
        self.image_id: str | None = None
        self.channel_id: str | None = None
        self._selected_images_visualise = {}
        self._image_gallery = ft.ListView()
        self._container_mask: ft.Container | None = None
        self._main_image: ft.Container | None = None
        self._interactive_viewer:FletExtendedInteractiveViewer | None = None
        self._mask_button:ft.IconButton | None = None
        self._edit_allowed = False
        self._edit_button:ft.IconButton | None = None
        self._slider_2d: ft.CupertinoSlidingSegmentedButton | None = None
        self._text_field_segmentation_channel: ft.TextField | None = None
        self._slider_2_5d:ft.Slider | None = None
        self._control_menu: ft.Container | None = None
        self._main_image_view: ft.Card | None = None
        self._ready = threading.Event()
        self._pipe_listener_running = True
        self._queue = None
        self._parent_conn, self._child_conn = None, None
        self._process_drawing_window = None
        self._window_image_id = ""
        self._window_bf_channel = ""
        self._window_channel_id = ""
        self._window_mask_path = ""
        self._thread = None
        Review._instances.append(self)

    @property
    def settings(self) -> ft.Stack:
        if self._settings is None:
            self._container_mask = ft.Container(
                ft.Image(src=r"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA\AAAFCAIAAAFe0wxPAAAAAElFTkSuQmCC",
                         fit=ft.ImageFit.SCALE_DOWN, ), visible=False, alignment=ft.alignment.center, width=632, height=632)

            self._main_image = ft.Container(
                ft.Image(src=r"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA\AAAFCAIAAAFe0wxPAAAAAElFTkSuQmCC",
                         fit=ft.ImageFit.SCALE_DOWN), alignment=ft.alignment.center, width=632, height=632)

            self._interactive_viewer = FletExtendedInteractiveViewer(
                content=ft.Stack([self._main_image, self._container_mask]), constrained=False, min_scale=0.1, width=632,
                height=632)
            zoom_value = 0.20
            self._mask_button = ft.IconButton(icon=ft.Icons.REMOVE_RED_EYE, icon_color=ft.Colors.BLACK12,
                                              style=ft.ButtonStyle(
                                                  shape=ft.RoundedRectangleBorder(radius=12), ),
                                              on_click=lambda e: self.show_mask(),
                                              tooltip="Show mask", hover_color=ft.Colors.WHITE12, disabled=True)
            self._edit_button = ft.IconButton(icon=ft.Icons.EDIT_SHARP, icon_color=ft.Colors.BLACK12,
                                  style=ft.ButtonStyle(
                                      shape=ft.RoundedRectangleBorder(radius=12),),
                                  on_click=lambda e: self.set_queue_drawing_window(),
                                  tooltip="Edit Mask", hover_color=ft.Colors.WHITE12,disabled=True)
            self._slider_2d = ft.CupertinoSlidingSegmentedButton(
                selected_index=0 if not self.user_2_5d else 1,
                thumb_color=ft.Colors.WHITE,
                bgcolor=ft.Colors.WHITE60,
                on_change=lambda e: self.slider_update(e),
                padding=ft.padding.symmetric(0, 0),
                controls=[
                    ft.Text("2D",color=ft.Colors.BLACK),
                    ft.Text("2.5D",color=ft.Colors.BLACK)
                ],
            )
            self._text_field_segmentation_channel = ft.TextField(
                border_color=ft.Colors.WHITE60,
                value=self.user_segmentation_channel,
                on_blur=lambda e: self.on_change_sc(e),
                tooltip="Segmentation channel",
                height=30, width=70, content_padding=ft.padding.symmetric(0, 5),
                fill_color=ft.Colors.WHITE38,
                filled=True,
                text_align=ft.TextAlign.CENTER,
                border_width=2,
                focused_border_color=ft.Colors.WHITE,
                text_style=ft.TextStyle(color=ft.Colors.BLACK,weight=ft.FontWeight.BOLD),
                cursor_color=ft.Colors.BLACK,
            )
            self._text_field_mask_suffix = ft.TextField(
                border_color=ft.Colors.WHITE60,
                value=self.user_mask_suffix,
                on_blur=lambda e: self.on_change_ms(e),
                tooltip="Mask suffix",
                height=30, width=70, content_padding=ft.padding.symmetric(0, 5),
                fill_color=ft.Colors.WHITE38,
                filled=True,
                text_align=ft.TextAlign.CENTER,
                border_width=2,
                focused_border_color=ft.Colors.WHITE,
                text_style=ft.TextStyle(color=ft.Colors.BLACK, weight=ft.FontWeight.BOLD),
                cursor_color=ft.Colors.BLACK,
                visible= False
            )
            self._slider_2_5d = ft.Slider(
                min=0, max=100, divisions=None, label="Slice: {value}", on_change=lambda e: self.slider_change(),
                opacity=1.0 if self.user_2_5d else 0.0, height=20,
                active_color=ft.Colors.WHITE60, thumb_color=ft.Colors.WHITE, disabled=True,
                animate_opacity= ft.Animation(duration=600, curve=ft.AnimationCurve.LINEAR_TO_EASE_OUT),
            )
            self._control_menu = ft.Container(ft.Container(ft.Row(
                [
                    ft.IconButton(icon=ft.Icons.ZOOM_IN, icon_color=ft.Colors.WHITE60,
                                  style=ft.ButtonStyle(
                                      shape=ft.RoundedRectangleBorder(radius=12), ),
                                  on_click=lambda e: self._interactive_viewer.zoom(1.0 + zoom_value), tooltip="Zoom in",
                                  hover_color=ft.Colors.WHITE12),
                    ft.IconButton(icon=ft.Icons.ZOOM_OUT, icon_color=ft.Colors.WHITE60,
                                  style=ft.ButtonStyle(
                                      shape=ft.RoundedRectangleBorder(radius=12), ),
                                  on_click=lambda e: self._interactive_viewer.zoom(1.0 - zoom_value), tooltip="Zoom out",
                                  hover_color=ft.Colors.WHITE12),
                    ft.IconButton(icon=ft.Icons.CROP_FREE, icon_color=ft.Colors.WHITE60,
                                  style=ft.ButtonStyle(
                                      shape=ft.RoundedRectangleBorder(radius=12), ),
                                  on_click=lambda e: self._interactive_viewer.reset(400),
                                  tooltip="Reset view", hover_color=ft.Colors.WHITE12),
                    self._text_field_segmentation_channel,
                    self._text_field_mask_suffix,
                    self._edit_button,
                    self._mask_button,
                    self._slider_2d,
                    ft.Container(
                        content=self._slider_2_5d,
                        theme=ft.Theme(
                            slider_theme=ft.SliderTheme(
                                value_indicator_text_style=ft.TextStyle(color=ft.Colors.BLACK, size=15,weight=ft.FontWeight.BOLD),
                            )
                        ),
                        dark_theme=ft.Theme(
                            slider_theme=ft.SliderTheme(
                                value_indicator_text_style=ft.TextStyle(color=ft.Colors.BLACK, size=15,weight=ft.FontWeight.BOLD),
                            )
                        ),
                    ),
                ], spacing=2,alignment=ft.MainAxisAlignment.CENTER,
            ), bgcolor=ft.Colors.BLUE_400, expand=True, border_radius=ft.border_radius.vertical(top=0, bottom=12),
            )
            )
            self._main_image_view = ft.Card(
                content=ft.Column([ft.Container(self._interactive_viewer, padding=ft.padding.only(top=10),
                                                alignment=ft.alignment.top_center), self._control_menu]),
                width=660, height=700,
                expand=True,
            )

            self._settings: ft.Stack = ft.Stack([ft.Row([ft.Column([ft.Row([
                self._main_image_view,
                ft.Card(content=ft.Container(self._image_gallery, width=600, height=700, expand=True, padding=20),
                        expand=True),
            ])
            ],
                alignment=ft.MainAxisAlignment.CENTER, )], alignment=ft.MainAxisAlignment.CENTER),])
        return self._settings

    def finished(self):
        self.outputs["mask_paths"].data = self.inputs["mask_paths"].data
        self._text_field_mask_suffix.visible = False
        self._text_field_mask_suffix.update()
        self._edit_allowed = False
        self._edit_button.icon_color = ft.Colors.BLACK12
        self._edit_button.disabled = True
        self._edit_button.update()
        self._pipe_listener_running = False
        self._queue.put("close")
        if self._process_drawing_window is not None and self._process_drawing_window.is_alive():
            self._process_drawing_window.join()
        self._child_conn.send("close")

        if self._thread is not None and self._thread.is_alive():
            self._thread.join()
        self._child_conn.close()
        self._parent_conn.close()
        self._queue = None
        self._parent_conn, self._child_conn = None, None
        self._process_drawing_window = None
        self._thread = None

    def run(self):
        self.event_manager.notify(ProgressEvent(percent=0, process=f"Preparing: starting"))
        self._ready.clear()
        self._pipe_listener_running = True
        self._queue = multiprocessing.Queue()
        parent_conn, child_conn = multiprocessing.Pipe()
        self._parent_conn, self._child_conn = parent_conn, child_conn
        self._process_drawing_window = self.start_drawing_window()
        self._thread = threading.Thread(target=self.child_conn_listener, daemon=True)
        self._thread.start()
        #reset
        self._window_image_id = ""
        self._window_bf_channel = ""
        self._window_channel_id = ""
        self._window_mask_path = ""
        self._icon_x = {}
        self._icon_check = {}
        self.image_id = None
        self.channel_id = None
        self._selected_images_visualise = {}
        self._image_gallery.clean()
        self._main_image.content.src=r"data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAUA\AAAFCAIAAAFe0wxPAAAAAElFTkSuQmCC"
        self._main_image.content.src_base64 = None
        self._main_image.update()
        self._container_mask.visible = False
        self._container_mask.update()
        self._edit_allowed = True
        self._ready.wait()
        self._ready.clear()
        self.event_manager.notify(ProgressEvent(percent=100, process=f"Preparing: finished"))
        self.event_manager.notify(ProgressEvent(percent=0, process=f"Loading Images: Starting"))
        src  = convert_tiffs_to_png_parallel(self.inputs["image_paths"].data)
        n_series = len(src)
        for iN,image_id in enumerate(src):
            cur_image_paths = src[image_id]

            self._selected_images_visualise[image_id] = {}
            for iN2,channel_id in enumerate(cur_image_paths):
                self._selected_images_visualise[image_id][channel_id] = ft.Container(
                    width=154,
                    height=154,
                    border=ft.border.all(4, ft.Colors.ORANGE_700),
                    alignment=ft.alignment.center,
                    visible=False,
                    padding=5
                )
            group_row = ft.Row(
                [
                    ft.Column(
                    [
                            ft.GestureDetector(
                                content=ft.Container(ft.Stack([ft.Image(
                                src_base64=cur_image_paths[channel_id],
                                height=150,
                                width=150,
                                fit=ft.ImageFit.CONTAIN
                                ),self._selected_images_visualise[image_id][channel_id]]),width=156,height=156),
                                on_tap=lambda e, img_id=image_id, c_id=channel_id: self.update_main_image(img_id, c_id),
                            ),
                            ft.Text(channel_id, size=10, text_align=ft.TextAlign.CENTER),
                        ],
                        alignment=ft.MainAxisAlignment.CENTER,
                        horizontal_alignment=ft.CrossAxisAlignment.CENTER,
                        spacing=5
                    )
                    for channel_id in cur_image_paths
                ],
                alignment=ft.MainAxisAlignment.START,
                spacing=10,
                scroll=ft.ScrollMode.AUTO,
            )
            self._icon_check[image_id] = ft.Icon(ft.Icons.CHECK, color=ft.Colors.GREEN, size=17, visible=False,
                                                tooltip="Mask is available")
            self._icon_x[image_id] = ft.Icon(ft.Icons.CLOSE, size=17, visible=True, tooltip="Mask not available")
            self.update_mask_check(image_id)
            self._image_gallery.controls.append(ft.Column([ft.Row(
            [ft.Text(f"{image_id}", weight=ft.FontWeight.BOLD, text_align=ft.TextAlign.CENTER), self._icon_check[image_id], self._icon_x[image_id]], spacing=2),
                                                      group_row], spacing=10, alignment=ft.MainAxisAlignment.CENTER))
            self.event_manager.notify(ProgressEvent(percent=int((iN+1) / n_series * 100), process=f"Loading Images: {iN+1}/{n_series}"))

        self.event_manager.notify(ProgressEvent(percent=100, process=f"Loading Images: Finished"))

        return True


    def update_mask_check(self, image_id):
        """
        Updates the symbol next to series number of image to a check or x, depending on if the corresponding image is available.
        Args:
            image_id: the id of the image to check mask availability
        """
        if self.inputs["mask_paths"].data is not None and image_id in self.inputs["mask_paths"].data and self.user_segmentation_channel in self.inputs["mask_paths"].data[image_id]:
            self._icon_check[image_id].visible = True
            self._icon_x[image_id].visible = False
        else:
            self._icon_check[image_id].visible = False
            self._icon_x[image_id].visible = True
        self._image_gallery.update()

    def update_all_masks_check(self):
        """
        Updates the symbol next to series number of image for every image_id in mask_paths.
        """
        if self.inputs["image_paths"].data is not None:
            for image_id in self.inputs["image_paths"].data:
                self.update_mask_check(image_id)

    def update_main_image(self,img_id,channel_id,on_click = True):
        """
        Method that handles what happens when the image is clicked or the main image need an update.
        """
        if on_click:
            if self.image_id is not None and self.image_id in self._selected_images_visualise:
                if self.channel_id is not None and self.channel_id in self._selected_images_visualise[self.image_id]:
                    self._selected_images_visualise[self.image_id][self.channel_id].visible = False
                    self._selected_images_visualise[self.image_id][self.channel_id].update()
        self.image_id = img_id
        self.channel_id = channel_id
        self._selected_images_visualise[img_id][channel_id].visible = True
        self._selected_images_visualise[img_id][channel_id].update()


        self._main_image.content.src_base64 = auto_adjust(self.inputs["image_paths"].data[img_id][channel_id], get_slice=int(self._slider_2_5d.value) if self.user_2_5d else -1)
        self._main_image.update()
        image = tifffile.imread(self.inputs["image_paths"].data[self.image_id][self.channel_id])
        self._text_field_mask_suffix.visible = not(self.inputs["mask_paths"].data is not None and self.image_id in self.inputs["mask_paths"].data and self.user_segmentation_channel in self.inputs["mask_paths"].data[self.image_id] and self.inputs["mask_paths"].data[self.image_id][self.user_segmentation_channel] is not None) and self._edit_allowed
        self._text_field_mask_suffix.update()

        if image.ndim == 3:
            if self._slider_2_5d.opacity == 1.0 and self._edit_allowed:
                self._edit_button.icon_color = ft.Colors.WHITE60
                self._edit_button.disabled = False
                self._edit_button.update()
            else:
                self._edit_button.icon_color = ft.Colors.BLACK12
                self._edit_button.disabled = True
                self._edit_button.update()
            self._slider_2_5d.value = 0 if image.shape[-2] - 1 < self._slider_2_5d.value else self._slider_2_5d.value
            self._slider_2_5d.max = image.shape[2] - 1
            self._slider_2_5d.divisions = image.shape[2] - 2
            self._slider_2_5d.disabled = False
            self._slider_2_5d.update()
        else:
            if self._edit_allowed:
                self._edit_button.icon_color = ft.Colors.WHITE60
                self._edit_button.disabled = False
                self._edit_button.update()
            self._slider_2_5d.value = 0
            self._slider_2_5d.max = 1
            self._slider_2_5d.divisions = None
            self._slider_2_5d.disabled = True
            self._slider_2_5d.update()

        if self.inputs["mask_paths"].data is not None and self.image_id in self.inputs["mask_paths"].data and self.user_segmentation_channel in self.inputs["mask_paths"].data[img_id]:
            if not self._container_mask.visible:
                self._mask_button.icon_color = ft.Colors.WHITE60
                self._mask_button.tooltip = "Show mask"
                self._mask_button.disabled = False
                self._mask_button.update()

            mask_data = np.load(Path(self.inputs["mask_paths"].data[self.image_id][self.user_segmentation_channel]), allow_pickle=True).item()

            mask= mask_data["masks"]
            outline = mask_data["outlines"]
            self._container_mask.content.src_base64 = self.convert_npy_to_canvas(mask,outline)
            self._container_mask.update()
        else:
            self._mask_button.tooltip = "Show mask"
            self._mask_button.icon_color = ft.Colors.BLACK12
            self._mask_button.disabled = True
            self._mask_button.update()
            self._container_mask.visible = False
            self._container_mask.update()

    def show_mask(self):
        self._container_mask.visible = not self._container_mask.visible
        self._container_mask.update()
        self._mask_button.icon_color = ft.Colors.WHITE if self._container_mask.visible else ft.Colors.WHITE60
        self._mask_button.tooltip="Hide mask" if self._container_mask.visible else "Show mask"
        self._mask_button.update()

    def convert_npy_to_canvas(self,mask, outline):
        """
        handles the conversion of the given file data

        Args:
            mask= the mask data stored in the numpy directory
            outline= the outline data stored in the numpy directory
        """
        buffer= BytesIO()

        if mask.ndim == 3:
            if self._slider_2_5d.opacity == 1.0:
                mask = np.take(mask, int(self._slider_2_5d.value), axis=0)
            else:
                mask = np.max(mask, axis=0)

        if outline.ndim == 3:
            if self._slider_2_5d.opacity == 1.0:
                outline = np.take(outline, int(self._slider_2_5d.value), axis=0)
            else:
                outline = np.max(outline, axis=0)

        image_mask = np.zeros(shape=(mask.shape[0], mask.shape[1], 4), dtype=np.uint8)
        r,g,b = self.mask_color
        image_mask[mask != 0] = (r, g, b, self.mask_opacity)
        r, g, b = self.outline_color
        image_mask[outline != 0] = (r, g, b, 255)
        im= Image.fromarray(image_mask).convert("RGBA")
        im.resize(size=(700,500))

        #saves the image as a image(base64)
        im.save(buffer, format="PNG")
        buffer.seek(0)
        image_base_64= base64.b64encode(buffer.getvalue()).decode('utf-8')

        #saves the created output image.
        return image_base_64


    def slider_update(self, e):
        if int(e.data) == 1:
            self._slider_2_5d.opacity = 1.0
            self.user_2_5d = True
        else:
            self._slider_2_5d.opacity = 0.0
            self.user_2_5d = False

        self.slider_change()
        self._slider_2_5d.update()
        self.event_manager.notify(OnPipelineChangeEvent("user_attr_change"))

    def slider_change(self):
        if self.image_id is not None:
            self.update_main_image(self.image_id, self.channel_id)

    def on_change_sc(self,e):
        if str(e.control.value) == "":
            self.settings.page.open(
                ft.SnackBar(
                    ft.Text(f"Segmentation channel must be not empty!",
                            color=ft.Colors.WHITE),
                    bgcolor=ft.Colors.RED))
            e.control.value = self.user_segmentation_channel
            self.settings.page.update()
            return
        self.user_segmentation_channel = str(e.control.value)
        self.update_all_masks_check()
        if self.image_id is not None:
            self.update_main_image(self.image_id, self.channel_id)
        self.event_manager.notify(OnPipelineChangeEvent("user_attr_change"))

    def on_change_ms(self,e):
        if str(e.control.value) == "":
            self.settings.page.open(
                ft.SnackBar(
                    ft.Text(f"Mask suffix must be not empty!",
                            color=ft.Colors.WHITE),
                    bgcolor=ft.Colors.RED))
            e.control.value = self.user_mask_suffix
            self.settings.page.update()
            return
        self.user_mask_suffix = str(e.control.value)
        self.event_manager.notify(OnPipelineChangeEvent("user_attr_change"))

    @classmethod
    def update_class(cls):
        for instance in cls._instances:
            if instance.image_id is not None:
                instance.update_main_image(instance.image_id, instance.channel_id)

    def destroy(self):
        self._instances.remove(self)
        super().destroy()


    def start_drawing_window(self):
        """
        Start the drawing window in multiprocessing.
        """
        process = multiprocessing.Process(target=open_qt_window,
                                args=(self._queue,self._child_conn))
        process.start()
        return process

    def set_queue_drawing_window(self):
        """
        Sets queue for drawing window with the current selected image and mask.
        """
        adjusted_image_path = os.path.join(
            Path(self.inputs["image_paths"].data[self.image_id][self.user_segmentation_channel]).parent,
            "adjusted_image.png")
        image_data = base64.b64decode(self._main_image.content.src_base64)
        buffer = BytesIO(image_data)
        image = Image.open(buffer)
        image.save(adjusted_image_path, format="PNG")

        if self._process_drawing_window is None or not self._process_drawing_window.is_alive(): #make sure that the process is running before putting new image in the queue
            if self._process_drawing_window is not None:
                try:
                    self._process_drawing_window.terminate()
                    self._process_drawing_window.join()
                except Exception as e:
                    self._settings.page.open(ft.SnackBar(ft.Text(f"Error while terminating process: {e}")))
            self._queue = multiprocessing.Queue()
            parent_conn, child_conn = multiprocessing.Pipe()
            self._parent_conn, self._child_conn = parent_conn, child_conn
            self._process_drawing_window = self.start_drawing_window()
        self._window_image_id = self.image_id
        self._window_bf_channel = self.user_segmentation_channel
        self._window_channel_id = self.channel_id

        if self._window_bf_channel in self.inputs["image_paths"].data[self.image_id]:#check if the bf has an image
            image_path = self.inputs["image_paths"].data[self.image_id][self._window_bf_channel]
            directory, filename = os.path.split(image_path)
            name, _ = os.path.splitext(filename)
            mask_file_name = f"{name}{self.user_mask_suffix}.npy"
            self._window_mask_path = os.path.join(directory, mask_file_name)

            i = len(self.image_id)
            j = len(name) - len(self._window_bf_channel)
            channel_prefix = name[i:j]
            self._queue.put((self.mask_color, self.outline_color,self.mask_opacity, self.user_segmentation_channel, self.inputs["mask_paths"].data, self._window_image_id, adjusted_image_path, self._window_mask_path,self._window_channel_id,channel_prefix,self._slider_2_5d.value if not self._slider_2_5d.disabled else None,self._slider_2_5d.max + 1 if not self._slider_2_5d.disabled else None))
        else:
            self._settings.page.open(ft.SnackBar(
                ft.Text(f"Selected bright-field channel {self._window_bf_channel} has no image!")))
            self._settings.page.update()

    def child_conn_listener(self):
        """
        Listener for the child connection.
        """
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        async def pipe_listener():
            while self._pipe_listener_running:
                data = await asyncio.to_thread(self._parent_conn.recv)
                split_data = data.split(".")
                if data == "close":
                    break
                elif split_data[0] == "new_mask":
                    if self._window_image_id not in self.inputs["mask_paths"].data:
                        self.inputs["mask_paths"].data[self._window_image_id] = {}
                    self.inputs["mask_paths"].data[self._window_image_id][
                        self._window_bf_channel] = self._window_mask_path
                    self.update_mask_check(split_data[1])
                    self._text_field_mask_suffix.visible = not (
                                self.image_id in self.inputs["mask_paths"].data and self.user_segmentation_channel in
                                self.inputs["mask_paths"].data[self.image_id] and
                                self.inputs["mask_paths"].data[self.image_id][
                                    self.user_segmentation_channel] is not None) and self._edit_allowed
                    self._text_field_mask_suffix.update()
                elif split_data[0] == "ready":
                    self._ready.set()
                else:
                    if self._window_image_id == self.image_id and self._window_bf_channel == self.user_segmentation_channel:
                        self.update_main_image(self.image_id,self.channel_id)
                    else:
                        self.reset_mask(self._window_image_id, self._window_bf_channel)
                        self._text_field_mask_suffix.visible = not (self.image_id in self.inputs[
                            "mask_paths"].data and self.user_segmentation_channel in self.inputs["mask_paths"].data[
                                                                        self.image_id] and
                                                                    self.inputs["mask_paths"].data[self.image_id][
                                                                        self.user_segmentation_channel] is not None) and self._edit_allowed
                        self._text_field_mask_suffix.update()
        try:
            loop.run_until_complete(pipe_listener())
        finally:
            loop.stop()
            loop.close()

    def reset_mask(self, image_id, segmentation_channel):
        if image_id in self.inputs["mask_paths"].data and segmentation_channel in self.inputs["mask_paths"].data[image_id]:
            del self.inputs["mask_paths"].data[image_id][segmentation_channel]

    def dismiss(self):
        if self._queue is not None:
            self._queue.put("hide")


import os
from pathlib import Path

from bigfish.detection.spot_detection import detect_spots
import numpy as np
import tifffile
from scipy.ndimage import binary_dilation

from cellsepi.backend.main_window.expert_mode.listener import ProgressEvent
from cellsepi.backend.main_window.expert_mode.module import *
from cellsepi.backend.main_window.expert_mode.pipeline import PipelineRunningException


class SpotDetectionModule(Module, ABC):
    _gui_config = ModuleGuiConfig("SpotDetection",Categories.SEGMENTATION,"This module handles spot detection in cells for each series on the given segmentation_channel with the big-fish python package.")
    def __init__(self, module_id: str) -> None:
        super().__init__(module_id)
        self.inputs = {
            "image_paths": InputPort("image_paths", dict),
            "mask_paths": InputPort("mask_paths", dict,opt=True),
        }
        self.outputs = {
            "mask_paths": OutputPort("mask_paths", dict),
        }
        self.user_remove_duplicate:bool = True
        self.user_use_threshold: bool = False
        self.user_use_log_kernel_and_minimum_distance:bool = False
        self.user_segmentation_channel: str = "2"
        self.user_mask_suffix: str = "_sdm" #spot detection mask
        self.user_mask_spot_radius_pixels: float = 3.0
        self.user_threshold: float = 355.0
        self.user_log_kernel_x_pixels: float = 1.456
        self.user_log_kernel_y_pixels: float = 1.456
        self.user_log_kernel_z_pixels: float = 1.167
        self.user_minimum_distance_x_pixels: float = 1.456
        self.user_minimum_distance_y_pixels: float = 1.456
        self.user_minimum_distance_z_pixels: float = 1.167
        self.user_voxel_size_x_nm: float = 103.0
        self.user_voxel_size_y_nm: float = 103.0
        self.user_voxel_size_z_nm: float = 300.0
        self.user_spot_radius_x_nm: float = 150.0
        self.user_spot_radius_y_nm: float = 150.0
        self.user_spot_radius_z_nm: float = 350.0

    @property
    def settings(self) -> ft.Stack|None:
        if self._settings is not None and self.on_change_user_use_log_kernel_and_minimum_distance() is None:
            self.on_change_user_use_log_kernel_and_minimum_distance = self.update_disable_kernel_distance
            self.on_change_user_use_threshold = self.update_disable_threshold
            self.update_disable_kernel_distance()
            self.update_disable_threshold()
        return self._settings

    def update_disable_threshold(self):
        if self.user_use_threshold:
            self.ref_user_threshold.current.disabled = False
        else:
            self.ref_user_threshold.current.disabled = True

    def update_disable_kernel_distance(self):
        if self.user_use_log_kernel_and_minimum_distance:
            self.ref_user_voxel_size_x_nm.current.disabled = True
            self.ref_user_voxel_size_y_nm.current.disabled = True
            self.ref_user_voxel_size_z_nm.current.disabled = True
            self.ref_user_spot_radius_x_nm.current.disabled = True
            self.ref_user_spot_radius_y_nm.current.disabled = True
            self.ref_user_spot_radius_z_nm.current.disabled = True
            self.ref_user_log_kernel_x_pixels.current.disabled = False
            self.ref_user_log_kernel_y_pixels.current.disabled = False
            self.ref_user_log_kernel_z_pixels.current.disabled = False
            self.ref_user_minimum_distance_x_pixels.current.disabled = False
            self.ref_user_minimum_distance_y_pixels.current.disabled = False
            self.ref_user_minimum_distance_z_pixels.current.disabled = False
        else:
            self.ref_user_voxel_size_x_nm.current.disabled = False
            self.ref_user_voxel_size_y_nm.current.disabled = False
            self.ref_user_voxel_size_z_nm.current.disabled = False
            self.ref_user_spot_radius_x_nm.current.disabled = False
            self.ref_user_spot_radius_y_nm.current.disabled = False
            self.ref_user_spot_radius_z_nm.current.disabled = False
            self.ref_user_log_kernel_x_pixels.current.disabled = True
            self.ref_user_log_kernel_y_pixels.current.disabled = True
            self.ref_user_log_kernel_z_pixels.current.disabled = True
            self.ref_user_minimum_distance_x_pixels.current.disabled = True
            self.ref_user_minimum_distance_y_pixels.current.disabled = True
            self.ref_user_minimum_distance_z_pixels.current.disabled = True

    def run(self):
        mask_paths = {}
        image_paths = self.inputs["image_paths"].data
        n_series = len(list(image_paths))
        self.event_manager.notify(ProgressEvent(percent=0, process=f"Spot detection: Starting"))
        for iN, image_id in enumerate(list(image_paths)):
            if self.user_segmentation_channel in image_paths[image_id] and os.path.isfile(self.inputs["image_paths"].data[image_id][self.user_segmentation_channel]):
                image_path = self.inputs["image_paths"].data[image_id][self.user_segmentation_channel]
                directory, filename = os.path.split(image_path)
                name, _ = os.path.splitext(filename)
                new_filename = f"{name}{self.user_mask_suffix}.npy"
                new_path = os.path.join(directory, new_filename)
                mask_paths[image_id] = {}
                mask_paths[image_id][self.user_segmentation_channel] = new_path
                image = tifffile.imread(image_path) #X,Y,Z
                if image.ndim == 3:
                    rna = np.transpose(image, (2,1,0)) #Z,Y,X for big-fish
                else:
                    rna = np.transpose(image, (1,0)) #Y,X for big-fish
                try:
                    spots, threshold = detect_spots(rna, remove_duplicate=self.user_remove_duplicate, threshold=None if not self.user_use_threshold else self.user_threshold, return_threshold=True,
                                                              voxel_size=(self.user_voxel_size_y_nm, self.user_voxel_size_x_nm) if image.ndim == 2 else (self.user_voxel_size_z_nm, self.user_voxel_size_y_nm, self.user_voxel_size_x_nm), spot_radius=(self.user_spot_radius_y_nm, self.user_spot_radius_x_nm) if image.ndim == 2 else (self.user_spot_radius_z_nm, self.user_spot_radius_y_nm, self.user_spot_radius_x_nm), log_kernel_size=None if not self.user_use_log_kernel_and_minimum_distance else (self.user_log_kernel_y_pixels, self.user_log_kernel_x_pixels) if image.ndim == 2 else (self.user_log_kernel_z_pixels, self.user_log_kernel_y_pixels, self.user_log_kernel_x_pixels),
                                                              minimum_distance=None if not self.user_use_log_kernel_and_minimum_distance else (self.user_minimum_distance_y_pixels, self.user_minimum_distance_x_pixels) if image.ndim == 2 else (self.user_minimum_distance_z_pixels, self.user_minimum_distance_y_pixels, self.user_minimum_distance_x_pixels))
                except Exception as e:
                    raise PipelineRunningException("Spot Detection Error",str(e))

                if image.ndim == 2:
                    empty_mask = {
                        "masks": np.zeros(image.shape, dtype=np.uint32),
                        "outlines": np.zeros(image.shape, dtype=np.uint32)
                    }
                else:
                    mask_shape = np.transpose(image, (2, 0, 1)).shape #Z,X,Y for cellpose masks
                    empty_mask = {
                        "masks": np.zeros(mask_shape, dtype=np.uint32),
                        "outlines": np.zeros(mask_shape, dtype=np.uint32)
                    }

                mask_seg = None
                if self.inputs["mask_paths"].data is not None:
                    if image_id in self.inputs["mask_paths"].data and self.user_segmentation_channel in self.inputs["mask_paths"].data[image_id]:
                        mask_seg = np.load(Path(self.inputs["mask_paths"].data[image_id][self.user_segmentation_channel]),allow_pickle=True).item()
                mask = create_spot_mask(spots, empty_mask,mask_seg, self.user_spot_radius_pixels)
                np.save(new_path, mask)

                self.event_manager.notify(ProgressEvent(percent=int((iN + 1) / n_series * 100),
                                                        process=f"Spot detection images: {iN + 1}/{n_series}"))

        self.outputs["mask_paths"].data = mask_paths
        self.event_manager.notify(ProgressEvent(percent=100, process=f"Spot detection: Finished"))


def create_spot_mask(spots:list, mask:dict,mask_seg:dict, radius:float, thickness:int=1):
    """
    Create a mask and outline for a list of spots.
    (3d spot input in (Z,Y,X) => 3d mask output in (Z,X,Y))
    Arguments:
        spots (list): List of spot coordinates. For 2D images: [(y, x), ...], for 3D images: [(z, y, x), ...].
        mask (dict): Dictionary with two numpy arrays:
                        - "masks": integer array for the spots.
                        - "outlines": integer array for spots outlines.
        mask_seg (dict): Dictionary with two numpy arrays:
                        - "masks": integer array for segmented cells.
                        - "outlines": integer array for segmented cells outlines.
        radius (float): Radius of each spot in pixels/voxels.
        thickness (int, optional (default=1)): Thickness of the outline in pixels/voxels.
    Returns:
        mask (dict): Dictionary with two numpy arrays:
                        - "masks": integer array for the spots.
                        - "outlines": integer array for spots outlines.
    """
    bool_3d = mask["masks"].ndim == 3
    masks = mask["masks"]
    masks_seg = mask_seg["masks"] if mask_seg is not None else None
    if not np.any(masks_seg):
        masks_seg = None
    outlines = mask["outlines"]

    for i, coordinates in enumerate(spots):
        spot_id = i + 1

        if not bool_3d:
            y, x = coordinates
            y, x = int(round(y)), int(round(x))
            h, w = masks.shape #Z,X,Y

            x_min = int(max(x - radius - thickness, 0))
            x_max = int(min(x + radius + thickness + 1, w))#end is exclusive
            y_min = int(max(y - radius - thickness, 0))
            y_max = int(min(y + radius + thickness + 1, h))#end is exclusive

            x_grid,y_grid  = np.ogrid[x_min:x_max,y_min:y_max]
            eu_dist = np.sqrt((x_grid - x) ** 2 + (y_grid - y) ** 2) #Euclidean distance

            #bb is bounding_box
            bb_masks = masks[x_min:x_max, y_min:y_max]
            bb_masks_seg = masks_seg[x_min:x_max, y_min:y_max] if masks_seg is not None else None
            bb_outlines = outlines[x_min:x_max, y_min:y_max]

            if bb_masks_seg is not None:
                bb_masks[(eu_dist <= radius) & (bb_masks_seg != 0)] = spot_id
                bin_mask = (bb_masks == spot_id)
                outline_mask = binary_dilation(bin_mask,iterations=thickness) & (~bin_mask)
                bb_outlines[outline_mask] = spot_id
            else:
                bb_masks[eu_dist <= radius] = spot_id
                bin_mask = (bb_masks == spot_id)
                outline_mask = binary_dilation(bin_mask,iterations=thickness) & (~bin_mask)
                bb_outlines[outline_mask] = spot_id
        else:
            z, y, x = coordinates
            z, y, x = int(round(z)), int(round(y)), int(round(x))
            d, h, w = masks.shape #Z,X,Y

            z_min = int(max(z - radius - thickness, 0))
            z_max = int(min(z + radius + thickness + 1, d))#end is exclusive
            x_min = int(max(x - radius - thickness, 0))
            x_max = int(min(x + radius + thickness + 1, w))#end is exclusive
            y_min = int(max(y - radius - thickness, 0))
            y_max = int(min(y + radius + thickness + 1, h))#end is exclusive

            z_grid, x_grid, y_grid  = np.ogrid[z_min:z_max,x_min:x_max,y_min:y_max]
            eu_dist = np.sqrt((z_grid - z) ** 2 + (x_grid - x) ** 2 + (y_grid - y) ** 2) #Euclidean distance

            #bb is bounding box
            bb_masks = masks[z_min:z_max, x_min:x_max, y_min:y_max]
            bb_masks_seg = masks_seg[z_min:z_max, x_min:x_max, y_min:y_max] if masks_seg is not None else None
            bb_outlines = outlines[z_min:z_max, x_min:x_max, y_min:y_max]

            if bb_masks_seg is not None:
                bb_masks[(eu_dist <= radius) & (bb_masks_seg != 0)] = spot_id
                bin_mask = (bb_masks == spot_id)
                outline_mask = binary_dilation(bin_mask,iterations=thickness) & (~bin_mask)
                bb_outlines[outline_mask] = spot_id
            else:
                bb_masks[eu_dist <= radius] = spot_id
                bin_mask = (bb_masks == spot_id)
                outline_mask = binary_dilation(bin_mask,iterations=thickness) & (~bin_mask)
                bb_outlines[outline_mask] = spot_id

    return mask




from pathlib import Path

from flet_extended_interactive_viewer import FletExtendedInteractiveViewer

from cellsepi.frontend.main_window.expert_mode.gui_pipeline import PipelineGUI
from cellsepi.frontend.main_window.expert_mode.expert_constants import *
from cellsepi.frontend.main_window.expert_mode.gui_pipeline_listener import PipelineChangeListener, ModuleExecutedListener, ModuleStartedListener, \
    ModuleProgressListener, ModuleErrorListener, DragAndDropListener, PipelinePauseListener, PipelineCancelListener, PipelineErrorListener
from cellsepi.backend.main_window.expert_mode.pipeline_storage import PipelineStorage

class Builder:
    def __init__(self,gui):
        self.gui = gui
        self.page = gui.page
        self.builder_page_stack = None
        self.pipeline_gui = PipelineGUI(self.page)
        self.pipeline_gui.interactive_view = None
        self.pipeline_running_event = None
        self.help_text =  ft.Container(
            content=ft.Column(
                controls=[
                    ft.Text(
                        "Move Modules",
                        size=50,
                        weight=ft.FontWeight.BOLD,
                        color=ft.Colors.GREY_500
                    ),
                    ft.Row([
                        ft.Text(
                            "here",
                            size=40,
                            italic=True,
                            weight=ft.FontWeight.BOLD,
                            color=ft.Colors.GREY_400
                        ),
                        ft.Icon(ft.Icons.CROP_FREE,size=50,color=ft.Colors.GREY_400),
                    ],
                    alignment=ft.MainAxisAlignment.CENTER,
                    ),
                ],
                spacing=2,
                alignment=ft.MainAxisAlignment.CENTER,
                horizontal_alignment=ft.CrossAxisAlignment.CENTER,
            )
            ,alignment=ft.alignment.center,
            width=self.page.width,
            height=self.page.height,
            animate_opacity= ft.Animation(duration=600, curve=ft.AnimationCurve.LINEAR_TO_EASE_OUT),
        )
        self.pipeline_storage = PipelineStorage(self.pipeline_gui)
        self.file_picker = ft.FilePicker(
            on_result=lambda a: self.on_select_file(a))
        self.file_saver = ft.FilePicker(
            on_result=lambda a: self.on_file_saved(a))
        self.page.overlay.extend([self.file_picker,self.file_saver])
        self.load_button = ft.IconButton(icon=ft.Icons.UPLOAD_FILE, on_click=lambda e: self.click_load_file(),
                                         icon_color=MAIN_ACTIVE_COLOR,
                                         style=ft.ButtonStyle(
                                             shape=ft.RoundedRectangleBorder(radius=12), ),
                                         tooltip="Load pipeline\n[Ctrl + L]", hover_color=ft.Colors.WHITE12)
        self.save_as_button = ft.IconButton(icon=ft.Icons.SAVE_AS_ROUNDED, on_click=lambda e: self.click_save_as_file(),
                                            icon_color=MAIN_ACTIVE_COLOR if len(self.pipeline_gui.modules) > 0 else ft.Colors.WHITE24, disabled=False if len(self.pipeline_gui.modules) > 0 else True,
                                            style=ft.ButtonStyle(
                                               shape=ft.RoundedRectangleBorder(radius=12), ),
                                            tooltip="Save as pipeline\n[Ctrl + Shift + S]", hover_color=ft.Colors.WHITE12)

        self.page.on_keyboard_event = lambda e: self.on_keyboard(e)
        self.save_button = ft.IconButton(icon=ft.Icons.SAVE_ROUNDED, on_click=lambda e: self.click_save_file(),
                                         icon_color=MAIN_ACTIVE_COLOR if self.pipeline_gui.pipeline_directory != "" else ft.Colors.WHITE24,
                                         disabled=False if self.pipeline_gui.pipeline_directory != "" else True,
                                         style=ft.ButtonStyle(
                                             shape=ft.RoundedRectangleBorder(radius=12), ),
                                         tooltip="Save pipeline\n[Ctrl + S]", hover_color=ft.Colors.WHITE12)
        self.run_menu_button = ft.IconButton(icon=ft.Icons.PLAY_ARROW, on_click=lambda e: self.run_menu_click(),
                                             icon_color=MAIN_ACTIVE_COLOR,
                                             style=ft.ButtonStyle(
                                             shape=ft.RoundedRectangleBorder(radius=12), ),
                                             tooltip="Show run menu\n[Ctrl + R]", hover_color=ft.Colors.WHITE12)
        self.delete_button = ft.IconButton(icon=ft.Icons.DELETE, on_click=lambda e: self.delete_button_click(), icon_color=MAIN_ACTIVE_COLOR,
                                           style=ft.ButtonStyle(
                                              shape=ft.RoundedRectangleBorder(radius=12),),
                                           tooltip="Show delete buttons\n[Ctrl + D]", hover_color=ft.Colors.WHITE12)
        self.port_button = ft.IconButton(icon=ft.Icons.VISIBILITY, on_click=lambda e: self.port_button_click(),
                                         icon_color=MAIN_ACTIVE_COLOR,
                                         style=ft.ButtonStyle(
                                               shape=ft.RoundedRectangleBorder(radius=12), ),
                                         tooltip="Show which ports get transferred\n[Ctrl + P]", hover_color=ft.Colors.WHITE12)

        self.left_tools = ft.Container(ft.Container(ft.Column(
                [
                    self.load_button, self.save_as_button,self.save_button,self.run_menu_button,self.delete_button,self.port_button
                ], tight=True,spacing=2
            ), bgcolor=MENU_COLOR, expand=True
            ),bgcolor=ft.Colors.TRANSPARENT,border_radius=ft.border_radius.all(10),
            bottom=BOTTOM_SPACING,left=SPACING_X,width=40,blur=10)


        self.start_button = ft.ElevatedButton(  # button to start the pipeline
            text="Start",
            icon=ft.Icons.PLAY_CIRCLE,
            tooltip="Start the pipeline",
            disabled=False if len(self.pipeline_gui.modules) > 0 else True,
            on_click=lambda e:self.run(),
            opacity=0.75,
        )

        self.resume_button = ft.ElevatedButton(  # button to resume the pipeline
            text="Resume",
            tooltip="Resume the pipeline",
            icon=ft.Icons.PLAY_CIRCLE,
            visible=False,
            on_click=lambda e: self.pipeline_gui.pipeline.resume(),
            opacity=0.75
        )


        self.cancel_button = ft.ElevatedButton(  # button to cancel the pipeline
            text="Cancel",
            tooltip="Cancel the pipeline",
            icon=ft.Icons.STOP_CIRCLE_ROUNDED,
            color=ft.Colors.RED,
            on_click=lambda e: self.cancel(),
            visible=False,
            opacity=0.75
        )
        self.progress_bar_module = ft.ProgressBar(value=0, width=220,bgcolor=ft.Colors.WHITE24,color=ft.Colors.BLUE_400)
        self.progress_pipeline = ft.ProgressRing(value=0,width=50,height=50,stroke_width=8,bgcolor=ft.Colors.WHITE24,color=ft.Colors.BLUE_400)
        self.progress_text = ft.Text(f"{self.pipeline_gui.pipeline.modules_executed}/{len(self.pipeline_gui.pipeline.modules)}", weight=ft.FontWeight.BOLD, tooltip="How many modules has been executed", color=MAIN_ACTIVE_COLOR)
        self.progress_stack = ft.Stack([self.progress_pipeline,ft.Container(self.progress_text,alignment=ft.alignment.center)],width=50,height=50,)
        self.progress_bar_module_text = ft.Text("0%", color=MAIN_ACTIVE_COLOR)
        self.progress_and_start = ft.Column([ft.Container(self.progress_stack,alignment=ft.alignment.center),
            ft.Container(
                content=ft.Stack([self.start_button, self.resume_button,self.cancel_button]),alignment=ft.alignment.center)],width=95,spacing=20
        )
        self.running_module = ft.Text("Module",color=ft.Colors.WHITE70,width=230,overflow=ft.TextOverflow.ELLIPSIS,max_lines=1,style=ft.TextThemeStyle.HEADLINE_SMALL)
        self.info_text = ft.Text("Idle, waiting for start.", color=MAIN_ACTIVE_COLOR, width=250, overflow=ft.TextOverflow.ELLIPSIS, max_lines=2)
        self.category_icon = ft.Icon(ft.Icons.CATEGORY_ROUNDED,color=ft.Colors.GREEN)
        self.run_infos = ft.Column([ft.Row([self.category_icon,self.running_module]),self.info_text])
        self.left_run_menu = ft.Column([
            self.run_infos,ft.Row([ft.Container(self.progress_bar_module),self.progress_bar_module_text],width=260),
        ],alignment=ft.MainAxisAlignment.SPACE_BETWEEN)

        self.zoom_menu_button = ft.IconButton(icon=ft.Icons.SEARCH, on_click=lambda e: self.zoom_menu_click(),
                                              icon_color=MAIN_ACTIVE_COLOR,
                                              style=ft.ButtonStyle(
                                                 shape=ft.RoundedRectangleBorder(radius=12), ),
                                              tooltip="Show zoom menu\n[Ctrl + M]", hover_color=ft.Colors.WHITE12)



        self.right_tools = ft.Container(ft.Container(ft.Column(
            [
                self.zoom_menu_button,
            ], tight=True, spacing=2
        ), bgcolor=MENU_COLOR, expand=True
        ), bgcolor=ft.Colors.TRANSPARENT, border_radius=ft.border_radius.all(10),
            bottom=BOTTOM_SPACING, right=SPACING_X+10, width=40, blur=10)

        self.interactive_view = None
        self.zoom_menu = ft.Container(ft.Container(ft.Row(
            [
                ft.IconButton(icon=ft.Icons.ZOOM_IN, icon_color=MAIN_ACTIVE_COLOR,
                              style=ft.ButtonStyle(
                                                 shape=ft.RoundedRectangleBorder(radius=12), ), on_click=lambda e: self.interactive_view.zoom(1.0+ZOOM_VALUE), tooltip="Zoom in\n[Ctrl + .]", hover_color=ft.Colors.WHITE12),
                ft.IconButton(icon=ft.Icons.ZOOM_OUT, icon_color=MAIN_ACTIVE_COLOR,
                              style=ft.ButtonStyle(
                                                 shape=ft.RoundedRectangleBorder(radius=12), ), on_click=lambda e: self.interactive_view.zoom(1.0-ZOOM_VALUE), tooltip="Zoom out\n[Ctrl + ,]", hover_color=ft.Colors.WHITE12),
                ft.IconButton(icon=ft.Icons.CROP_FREE, icon_color=MAIN_ACTIVE_COLOR,
                              style=ft.ButtonStyle(
                               shape=ft.RoundedRectangleBorder(radius=12), ),
                              on_click=lambda e: self.interactive_view.reset(400), tooltip="Reset view\n[Ctrl + -]", hover_color=ft.Colors.WHITE12),
            ], spacing=2
        ), bgcolor=MENU_COLOR, expand=True
        ), bgcolor=ft.Colors.TRANSPARENT, border_radius=ft.border_radius.all(10),
            bottom=BOTTOM_SPACING, right=self.right_tools.right + self.right_tools.width + 5, blur=10, opacity=0,
            animate_opacity=ft.Animation(duration=300, curve=ft.AnimationCurve.LINEAR_TO_EASE_OUT),
            animate=ft.Animation(duration=300, curve=ft.AnimationCurve.LINEAR_TO_EASE_OUT),
        )

        self.run_menu = ft.Container(ft.Container(ft.Row(
            [
                ft.Container(self.left_run_menu,padding=10),ft.VerticalDivider(), ft.Column([ft.Row([ft.Container(self.progress_and_start,padding=10)],alignment=ft.MainAxisAlignment.CENTER)],alignment=ft.MainAxisAlignment.CENTER),
            ], spacing=2
        ), bgcolor=MENU_COLOR, expand=True, padding=10
        ), bgcolor=ft.Colors.TRANSPARENT, border_radius=ft.border_radius.all(10),width=0,height=150,
            bottom=BOTTOM_SPACING, left=self.left_tools.left + self.left_tools.width + 5,blur=10,opacity=0,
            animate_opacity=ft.Animation(duration=300, curve=ft.AnimationCurve.LINEAR_TO_EASE_OUT),
            animate=ft.Animation(duration=300, curve=ft.AnimationCurve.LINEAR_TO_EASE_OUT),
            )
        self.work_area = None
        self.setup()
        self.page_forward = ft.IconButton(icon=ft.Icons.CHEVRON_RIGHT_SHARP, on_click=lambda e: self.press_page_forward(),
                                          icon_color=MAIN_ACTIVE_COLOR,
                                          style=ft.ButtonStyle(
                                         shape=ft.RoundedRectangleBorder(radius=12), ),
                                          tooltip="Get to the next page\n[Ctrl + E]", hover_color=ft.Colors.WHITE12)
        self.page_backward = ft.IconButton(icon=ft.Icons.CHEVRON_LEFT_SHARP, on_click=lambda e: self.press_page_backward(),
                                           icon_color=ft.Colors.WHITE24,
                                           style=ft.ButtonStyle(
                                           shape=ft.RoundedRectangleBorder(radius=12), ), disabled=True,
                                           tooltip="Return to the last page\n[Ctrl + Q]", hover_color=ft.Colors.WHITE12,)
        self.pipeline_gui.build_show_room(self.builder_page_stack)
        self.pipeline_gui.interactive_view = self.interactive_view
        self.switch_pages = ft.Container(ft.Container(ft.Row(
                    [
                        self.page_backward, self.page_forward,
                    ], tight=True,spacing=2
                ), bgcolor=MENU_COLOR, expand=True, height=40
                ), bgcolor=ft.Colors.TRANSPARENT, border_radius=ft.border_radius.all(10),
                    top=self.pipeline_gui.show_room_container.top + self.pipeline_gui.show_room_container.height + 5,
                    left=self.pipeline_gui.show_room_container.left,blur=10,visible=True if self.pipeline_gui.show_room_max_page_number > 1 else False)
        self.builder_page_stack.controls.insert(1, self.switch_pages)
        self.add_all_listeners()

    def cancel(self):
        """
        To cancel the pipeline after the currently executed module is finished.
        """
        if self.cancel_button.visible:
            self.running_module.value = f"Pipeline"
            self.running_module.update()
            self.category_icon.color = ft.Colors.RED
            self.category_icon.update()
            self.cancel_button.disabled = True
            self.cancel_button.color = None
            self.cancel_button.update()
            self.info_text.value = ""
            self.info_text.spans = [
                ft.TextSpan("Canceling: ", style=ft.TextStyle(weight=ft.FontWeight.BOLD, color=ft.Colors.RED)),
                ft.TextSpan("...", style=ft.TextStyle(color=ft.Colors.WHITE60)), ]
            self.info_text.update()
            self.page.update()

        self.pipeline_gui.pipeline.cancel()

    def run(self,ignore_check=False):
        """
        To run the pipeline.

        Arguments:
            ignore_check (bool): if a warning should appear when you have not satisfied modules in the pipeline.
        """
        show_room_module_ids = [m.module_id for m in self.pipeline_gui.show_room_modules]
        if not ignore_check and not self.pipeline_gui.pipeline.check_pipeline_runnable(show_room_module_ids):
            def dismiss_dialog(e):
                cupertino_alert_dialog.open = False
                e.control.page.update()
                for mod in self.pipeline_gui.modules.values():
                    if not self.pipeline_gui.pipeline.check_module_satisfied(mod.module_id):
                        if not mod.show_ports:
                            mod.ports_in_out_clicked()
            def dismiss_dialog_ignore(e):
                cupertino_alert_dialog.open = False
                e.control.page.update()
                self.run(True)
            cupertino_alert_dialog = ft.CupertinoAlertDialog(
                title=ft.Text("Mandatory Input Warning"),
                content=ft.Text("Not all mandatory inputs are satisfied."),
                actions=[
                    ft.CupertinoDialogAction(
                        "Change modules",is_default_action=True, on_click=dismiss_dialog
                    ),
                    ft.CupertinoDialogAction(text="Skip modules", is_destructive_action=True, on_click=dismiss_dialog_ignore),
                ],
            )
            self.page.overlay.append(cupertino_alert_dialog)
            cupertino_alert_dialog.open = True
            self.page.update()
            return
        self.info_text.spans = []
        self.info_text.value = "Idle, waiting for start."
        self.info_text.update()
        self.start_button.visible = False
        self.start_button.update()
        self.cancel_button.visible = True
        self.cancel_button.update()
        self.load_button.disabled = True
        self.load_button.icon_color = ft.Colors.WHITE24
        self.load_button.update()
        for module in self.pipeline_gui.modules.values():
            self.pipeline_gui.lines_gui.update_delete_buttons(module,True)
            module.waiting_button.visible = True
            module.delete_button.visible = False
            module.waiting_button.update()
            module.delete_button.update()
            module.disable_tools()
            module.error_stack.visible = False
            module.error_stack.update()
            module.check_warning()
        self.gui.training_environment.disable_switch_environment()
        self.update_modules_executed(reset=True)

        self.pipeline_gui.pipeline.run(show_room_module_ids)
        if len(self.pipeline_gui.pipeline.modules) - len(ModuleType) * 2 != self.pipeline_gui.module_count or self.pipeline_gui.module_count != self.pipeline_gui.pipeline.modules_executed:
            self.update_modules_executed(reset=True)
        self.start_button.visible = True
        self.start_button.update()
        self.cancel_button.visible = False
        self.cancel_button.update()
        self.load_button.disabled = False
        self.load_button.icon_color = MAIN_ACTIVE_COLOR
        self.load_button.update()
        self.gui.training_environment.enable_switch_environment()
        if self.pipeline_running_event is not None:
            self.pipeline_running_event.set()

    def add_all_listeners(self):
        """
        Adds all listeners to this pipeline.
        """
        pipeline_change_listener = PipelineChangeListener(self)
        self.pipeline_gui.pipeline.event_manager.subscribe(listener=pipeline_change_listener)
        module_executed_listener =ModuleExecutedListener(self)
        self.pipeline_gui.pipeline.event_manager.subscribe(listener=module_executed_listener)
        module_started_listener =ModuleStartedListener(self)
        self.pipeline_gui.pipeline.event_manager.subscribe(listener=module_started_listener)
        module_progress_listener =ModuleProgressListener(self)
        self.pipeline_gui.pipeline.event_manager.subscribe(listener=module_progress_listener)
        module_error_listener =ModuleErrorListener(self)
        self.pipeline_gui.pipeline.event_manager.subscribe(listener=module_error_listener)
        drag_and_drop_listener =DragAndDropListener(self)
        self.pipeline_gui.pipeline.event_manager.subscribe(listener=drag_and_drop_listener)
        pipeline_pause_listener =PipelinePauseListener(self)
        self.pipeline_gui.pipeline.event_manager.subscribe(listener=pipeline_pause_listener)
        pipeline_cancel_listener =PipelineCancelListener(self)
        self.pipeline_gui.pipeline.event_manager.subscribe(listener=pipeline_cancel_listener)
        pipeline_error_listener =PipelineErrorListener(self)
        self.pipeline_gui.pipeline.event_manager.subscribe(listener=pipeline_error_listener)

    def on_keyboard(self,e: ft.KeyboardEvent):
        """
        All Keyboard shortcuts of the ExpertMode.
        Only working when ExpertMode visible
        """
        if not self.gui.ref_builder_environment.current.visible:
            return
        if e.shift and e.ctrl and e.key == "S" and not e.alt and not e.meta:
            if not self.save_as_button.disabled:
                self.click_save_as_file()
        if e.ctrl and e.key == "S" and not e.alt and not e.shift and not e.meta:
            if not self.save_button.disabled:
                self.click_save_file()
        if e.ctrl and e.key == "L" and not e.alt and not e.shift and not e.meta:
            if not self.load_button.disabled:
                self.click_load_file()
        if e.ctrl and e.key == "R" and not e.alt and not e.shift and not e.meta:
            self.run_menu_click()
        if e.ctrl and e.key == "D" and not e.alt and not e.shift and not e.meta:
            self.delete_button_click()
        if e.ctrl and e.key == "P" and not e.alt and not e.shift and not e.meta:
            self.port_button_click()
        if e.ctrl and e.key == "Q" and not e.alt and not e.shift and not e.meta:
            if not self.page_backward.disabled:
                self.press_page_backward()
        if e.ctrl and e.key == "E" and not e.alt and not e.shift and not e.meta:
            if not self.page_forward.disabled:
                self.press_page_forward()
        if e.ctrl and e.key == "M" and not e.alt and not e.shift and not e.meta:
            self.zoom_menu_click()
        if e.ctrl and e.key == "." and not e.alt and not e.shift and not e.meta:
            self.interactive_view.zoom(1.0+ZOOM_VALUE)
        if e.ctrl and e.key == "," and not e.alt and not e.shift and not e.meta:
            self.interactive_view.zoom(1.0-ZOOM_VALUE)
        if e.ctrl and e.key == "-" and not e.alt and not e.shift and not e.meta:
            self.interactive_view.reset(400)

    def update_modules_executed(self,reset:bool=False):
        """
        Updates the gui of the run menu how many modules were executed.

        Arguments:
            reset (bool): To reset the pipeline modules executed to 0, because the pipeline changed.
        """
        if reset:
            self.pipeline_gui.pipeline.modules_executed = 0
        current =self.pipeline_gui.pipeline.modules_executed
        if not self.pipeline_gui.pipeline.running:
            total = len(self.pipeline_gui.pipeline.modules) - len(ModuleType) * 2
            self.progress_pipeline.value = (current / total) if total > 0 else 0
            self.pipeline_gui.module_count = total
            self.progress_text.value = f"{current}/{total}"
        else:
            self.progress_pipeline.value = (current / self.pipeline_gui.module_count) if self.pipeline_gui.module_count > 0 else 0
            self.progress_text.value = f"{current}/{self.pipeline_gui.module_count}"
        self.progress_text.update()
        self.page.update()

    def click_load_file(self):
        """
        Called when clicked a file should be loaded.
        """
        self.file_picker.pick_files(file_type=ft.FilePickerFileType.CUSTOM, allowed_extensions=["csp"],
                                    allow_multiple=False)
        self.load_button.icon_color = ft.Colors.BLUE_400
        self.load_button.update()

    def click_save_as_file(self):
        """
        Called when clicked to save a file is at a specific location.
        """
        self.file_saver.save_file(file_type=ft.FilePickerFileType.CUSTOM, allowed_extensions=["csp"],
                             dialog_title="Save Pipeline", file_name=self.pipeline_gui.pipeline_name,
                             initial_directory=self.pipeline_gui.pipeline_directory)
        self.save_as_button.icon_color = ft.Colors.BLUE_400
        self.save_as_button.update()

    def click_save_file(self):
        """
        Called when clicked a file should be saved.
        """
        self.save_button.icon_color = ft.Colors.BLUE_400
        self.save_button.update()
        path = self.pipeline_storage.save_pipeline()
        self.pipeline_gui.page.open(
            ft.SnackBar(ft.Text(f"Pipeline saved at {path}", color=ft.Colors.WHITE), bgcolor=ft.Colors.GREEN))
        self.pipeline_gui.page.update()

        self.page.title = f"CellSePi - {self.pipeline_gui.pipeline_name}"
        self.save_button.icon_color = ft.Colors.WHITE24
        self.save_button.disabled = True
        self.page.update()

    def on_select_file(self, e):
        """
        Handles if a file is selected.
        """
        if e.files is not None:
            if not self.pipeline_storage.check_saved():
                def cancel_dialog(a):
                    cupertino_alert_dialog.open = False
                    a.control.page.update()

                def ok_dialog(a):
                    cupertino_alert_dialog.open = False
                    a.control.page.update()
                    if self.pipeline_gui.pipeline.running:
                        self.pipeline_gui.page.open(
                            ft.SnackBar(
                                ft.Text(f"Failed to load pipeline: a previous pipeline execution is still active!",
                                        color=ft.Colors.WHITE),
                                bgcolor=ft.Colors.RED))
                        self.pipeline_gui.page.update()
                        return
                    try:
                        self.pipeline_storage.load_pipeline(e.files[0].path)
                        self.pipeline_gui.reset()
                        self.pipeline_gui.load_pipeline()
                    except Exception as exception2:
                        self.pipeline_gui.page.open(
                            ft.SnackBar(
                                ft.Text(f"Failed to load pipeline: {exception2}",
                                        color=ft.Colors.WHITE),
                                bgcolor=ft.Colors.RED))
                        self.pipeline_gui.page.update()

                cupertino_alert_dialog = ft.CupertinoAlertDialog(
                    title=ft.Text("Unsaved Changes"),
                    content=ft.Text("Loading will discard any unsaved changes to the currently opened pipeline."),
                    actions=[
                        ft.CupertinoDialogAction(
                            "Cancel",is_default_action=True, on_click=cancel_dialog
                        ),
                        ft.CupertinoDialogAction(text="Ok", is_destructive_action=True, on_click=ok_dialog),
                    ],
                )
                self.page.overlay.append(cupertino_alert_dialog)
                cupertino_alert_dialog.open = True
                self.page.update()
                self.load_button.icon_color = MAIN_ACTIVE_COLOR
                self.load_button.update()
                return
            else:
                if self.pipeline_gui.pipeline.running:
                    self.pipeline_gui.page.open(
                        ft.SnackBar(ft.Text(f"Failed to load pipeline: a previous pipeline execution is still active!", color=ft.Colors.WHITE),
                                    bgcolor=ft.Colors.RED))
                    self.pipeline_gui.page.update()
                    return
                try:
                    self.pipeline_storage.load_pipeline(e.files[0].path)
                    self.pipeline_gui.reset()
                    self.pipeline_gui.load_pipeline()
                except Exception as exception1:
                    self.pipeline_gui.page.open(
                        ft.SnackBar(
                            ft.Text(f"Failed to load pipeline: {exception1}",
                                    color=ft.Colors.WHITE),
                            bgcolor=ft.Colors.RED))
                    self.pipeline_gui.page.update()

        self.load_button.icon_color = MAIN_ACTIVE_COLOR
        self.load_button.update()


    def on_file_saved(self, e):
        """
        Handles if a file gets saved.
        """
        if e.path is not None:
            if Path(e.path).suffix == "":
                e.path = e.path + ".csp"
            if Path(e.path).suffix != ".csp":
                self.pipeline_gui.page.open(ft.SnackBar(ft.Text(f"Pipeline name must have .csp suffix!",color=ft.Colors.WHITE),bgcolor=ft.Colors.RED))
                self.pipeline_gui.page.update()
                self.page.title = f"CellSePi - {self.pipeline_gui.pipeline_name}*"
                self.save_as_button.icon_color = MAIN_ACTIVE_COLOR
                self.page.update()
                return
            self.pipeline_storage.save_as_pipeline(e.path)
            self.pipeline_gui.page.open(ft.SnackBar(ft.Text(f"Pipeline saved at {e.path}",color=ft.Colors.WHITE),bgcolor=ft.Colors.GREEN))
            self.pipeline_gui.page.update()
            self.page.title = f"CellSePi - {self.pipeline_gui.pipeline_name}"
            self.save_button.icon_color = ft.Colors.WHITE24
            self.save_button.disabled = True
            self.page.update()

        self.save_as_button.icon_color = MAIN_ACTIVE_COLOR
        self.save_as_button.update()


    def press_page_forward(self):
        """
        Called when clicked to load the next page.
        """
        self.pipeline_gui.change_show_room_page(self.pipeline_gui.show_room_page_number + 1)
        if self.pipeline_gui.show_room_page_number > 0:
            self.page_backward.icon_color = MAIN_ACTIVE_COLOR
            self.page_backward.disabled = False
            self.page_backward.update()
        if self.pipeline_gui.show_room_page_number >= self.pipeline_gui.show_room_max_page_number-1:
            self.page_forward.icon_color = ft.Colors.WHITE24
            self.page_forward.disabled = True
            self.page_forward.update()

    def press_page_backward(self):
        """
        Called when clicked to load the previous page.
        """
        self.pipeline_gui.change_show_room_page(self.pipeline_gui.show_room_page_number - 1)
        if self.pipeline_gui.show_room_page_number == 0:
            self.page_backward.icon_color = ft.Colors.WHITE24
            self.page_backward.disabled = True
            self.page_backward.update()
        if self.pipeline_gui.show_room_page_number < self.pipeline_gui.show_room_max_page_number-1:
            self.page_forward.icon_color = MAIN_ACTIVE_COLOR
            self.page_forward.disabled = False
            self.page_forward.update()

    def run_menu_click(self):
        """
        Called when the run menu button got clicked.
        """
        if self.run_menu.opacity==1:
            self.run_menu_button.icon_color = MAIN_ACTIVE_COLOR
            self.run_menu_button.tooltip = f"Show run menu\n[Ctrl + R]"
            self.run_menu_button.update()
            self.run_menu.width = 0
            self.run_menu.opacity = 0
            self.run_menu.update()
        else:
            self.run_menu_button.icon_color = ft.Colors.BLUE_400
            self.run_menu_button.tooltip = f"Hide run menu\n[Ctrl + R]"
            self.run_menu_button.update()
            self.run_menu.width = 430
            self.run_menu.opacity = 1
            self.run_menu.update()

    def zoom_menu_click(self):
        """
        Called when the zoom menu button got clicked.
        """
        if self.zoom_menu.opacity==1:
            self.zoom_menu_button.icon_color = MAIN_ACTIVE_COLOR
            self.zoom_menu_button.tooltip = f"Show zoom menu\n[Ctrl + M]"
            self.zoom_menu_button.update()
            self.zoom_menu.width = 0
            self.zoom_menu.opacity = 0
            self.zoom_menu.update()
        else:
            self.zoom_menu_button.icon_color = ft.Colors.BLUE_400
            self.zoom_menu_button.tooltip = f"Hide zoom menu\n[Ctrl + M]"
            self.zoom_menu_button.update()
            self.zoom_menu.width = 122
            self.zoom_menu.opacity = 1
            self.zoom_menu.update()

    def delete_button_click(self):
        """
        Called when the delete button is clicked and toggels between all delete buttons be visible or be hidden.
        """
        if self.pipeline_gui.show_delete_button:
            self.delete_button.icon_color = MAIN_ACTIVE_COLOR
            self.delete_button.tooltip = f"Show delete buttons\n[Ctrl + D]"
            self.pipeline_gui.show_delete_button = False
            self.pipeline_gui.lines_gui.update_all()
        else:
            self.delete_button.icon_color = ft.Colors.BLUE_400
            self.delete_button.tooltip = f"Hide delete buttons\n[Ctrl + D]"
            self.pipeline_gui.show_delete_button = True
            if self.pipeline_gui.show_ports:
                self.port_button_click()
            self.pipeline_gui.lines_gui.update_all()

        self.delete_button.update()

    def port_button_click(self):
        """
        Called when the port button is clicked and toggels between all port text's be visible or be hidden.
        """
        if self.pipeline_gui.show_ports:
            self.port_button.icon_color = MAIN_ACTIVE_COLOR
            self.port_button.tooltip = f"Show which ports get transferred\n[Ctrl + P]"
            self.pipeline_gui.show_ports = False
            self.pipeline_gui.lines_gui.update_all()
        else:
            self.port_button.icon_color = ft.Colors.BLUE_400
            self.port_button.tooltip = f"Hide which ports get transferred\n[Ctrl + P]"
            self.pipeline_gui.show_ports = True
            if self.pipeline_gui.show_delete_button:
                self.delete_button_click()
            self.pipeline_gui.lines_gui.update_all()

        self.port_button.update()


    def setup(self):
        """
        Setup all the GUI elements.
        """
        self.work_area = ft.Stack([self.help_text,ft.Container(
            content=self.pipeline_gui,
            width=10000,
            height=10000,
            bgcolor=ft.Colors.TRANSPARENT,
        )])

        self.interactive_view = FletExtendedInteractiveViewer(content=self.work_area, constrained=False,
                                                              height=self.page.window.height,
                                                              width=self.page.window.width, scale_enabled=False,)

        def on_resize(e: ft.WindowResizeEvent):
            """
            Called when the resize-event is triggered.
            Updates all relevant GUI elements.
            """
            self.interactive_view.height = e.height-20
            self.interactive_view.width = e.width
            self.interactive_view.update()
            self.pipeline_gui.update_show_room()
            self.help_text.height = e.height
            self.help_text.width = e.width
            self.help_text.update()

        self.page.on_resized = on_resize

        self.builder_page_stack = ft.Stack([
                self.interactive_view,
                self.left_tools,
                self.right_tools,
                self.run_menu,
                self.zoom_menu,
             ]
            )



import flet as ft
import asyncio

class ExpertEnvironment(ft.Container):
    def __init__(self, gui):
        super().__init__()
        self.gui = gui
        self.text = ft.Text("Go To Expert Mode")
        self.button_event = ft.PopupMenuItem(
            content=ft.Row(
                [
                    ft.Icon(ft.Icons.EXIT_TO_APP),
                    self.text
                ]
            ),
            on_click=lambda e: self.change_environment(e),
        )
        self.switch_icon = ft.Icon(ft.Icons.REBASE_EDIT)
        self.button_expert_environment_menu = ft.PopupMenuButton(
            items=[self.button_event],
            content=self.switch_icon,
            tooltip="Expert Mode",
            on_open=lambda _: self.text.update(),
        )
        self.content = self.button_expert_environment_menu
        self.padding = 10
        self.alignment = ft.alignment.top_right
        self.old_view: (float, float,float) = (0.0, 0.0, 1.0)

    def change_environment(self, e):
        """
        Changing between modes.
        """
        if self.text.value == "Go To Expert Mode":
            self.go_to_expert_environment(e)
        else:
            self.old_view = self.gui.builder_environment.interactive_view.get_transformation_data()
            self.gui.ref_builder_environment.current.visible = False
            self.gui.ref_seg_environment.current.visible = True
            self.gui.ref_gallery_environment.current.visible = True
            self.page.title = "CellSePi"
            self.gui.page.update()
            self.text.value = "Go To Expert Mode"

    def go_to_expert_environment(self, e):
        """
        Switching to expert mode.
        """
        self.gui.ref_builder_environment.current.visible = True
        self.gui.ref_gallery_environment.current.visible = False
        self.gui.ref_training_environment.current.visible = False
        self.gui.ref_seg_environment.current.visible = False
        star = "*" if not self.gui.builder_environment.pipeline_storage.check_saved() else ""
        self.page.title = f"CellSePi - {self.gui.builder_environment.pipeline_gui.pipeline_name}{star}"
        self.gui.page.update()
        self.text.value = "Exit Expert Mode"
        self.gui.training_environment.text.value = "Go To Training"
        self.page.run_task(self._update_view)

    async def _update_view(self):
        """
        Loads the view of the expert environment, so its view has the orginal state when leaving the environment.
        """
        await asyncio.sleep(0.1)
        self.gui.builder_environment.interactive_view.set_transformation_data(self.old_view[0], self.old_view[1],self.old_view[2],300)
        self.gui.builder_environment.interactive_view.update()




import base64
import os
import pathlib
import platform
import shutil
import stat
from concurrent.futures import ThreadPoolExecutor
from enum import Enum, auto
from io import BytesIO

import numpy as np
from PIL import Image
from bioio import BioImage
import bioio_lif
from tifffile import tifffile

from cellsepi.backend.main_window.expert_mode.event_manager import *


def listdir(directory):
    dir_list = [directory / elem for elem in os.listdir(directory)]
    return dir_list


def organize_files(files, channel_prefix, mask_suffix=""):
    id_to_file = {}
    for file in files:
        if channel_prefix in file.name:
            image_id, channel_id = file.stem.replace(mask_suffix, "").split(channel_prefix)
            if image_id not in id_to_file:
                id_to_file[image_id] = {}

            if channel_id in id_to_file[image_id]:
                raise Exception(
                    f"""The directory already includes a file with the same image and channel ids.
                                Image Id: {image_id}
                                Channel Id: {channel_id}
                                Path: {file}""")

            id_to_file[image_id][channel_id] = file

    #sorting the Channel IDs
    for image_id in id_to_file:
        id_to_file[image_id] = dict(sorted(id_to_file[image_id].items()))
    #sorting the Image IDs
    id_to_file = dict(sorted(id_to_file.items()))
    return id_to_file

class ReturnTypePath(Enum):
    IMAGE_PATHS = auto()
    MASK_PATHS = auto()
    BOTH_PATHS = auto()

def load_directory(directory, channel_prefix=None, mask_suffix=None,return_type: ReturnTypePath=ReturnTypePath.BOTH_PATHS,event_manager: EventManager=None):
    assert directory is not None

    total_steps = 4 if return_type == ReturnTypePath.BOTH_PATHS else 3
    step = 0

    def notifier(process:str):
        nonlocal step
        step += 1
        if event_manager is not None:
            event_manager.notify(event=ProgressEvent(int(step/total_steps*100), process=process))

    if channel_prefix is None:
        channel_prefix = "c"

    if mask_suffix is None:
        mask_suffix = "_seg"

    if event_manager is not None:
        event_manager.notify(event=ProgressEvent(0, process="Organizing: Listing Directory"))

    names = os.listdir(directory)
    paths = [directory / name for name in names]
    file_paths = [path for path in paths if path.is_file()]


    notifier("Organizing: Filtering Directory for Images")
    tiff_files = [path for path in file_paths if path.suffix == ".tif" or path.suffix == ".tiff"]

    match return_type:
            case ReturnTypePath.IMAGE_PATHS:
                notifier( "Organizing: Image Files")
                id_to_image = organize_files(tiff_files, channel_prefix=channel_prefix)
                notifier( "Finished Organizing Files")
                return id_to_image
            case ReturnTypePath.MASK_PATHS:
                notifier( "Organizing: Mask Files")
                mask_files = [path for path in file_paths if path.suffix == ".npy" and path.stem.endswith(mask_suffix)]
                id_to_mask = organize_files(mask_files, channel_prefix=channel_prefix, mask_suffix=mask_suffix)
                notifier( "Finished Organizing Files")
                return id_to_mask
            case ReturnTypePath.BOTH_PATHS:
                notifier( "Organizing: Image Files")
                id_to_image = organize_files(tiff_files, channel_prefix=channel_prefix)
                notifier( "Organizing: Mask Files")
                mask_files = [path for path in file_paths if path.suffix == ".npy" and path.stem.endswith(mask_suffix)]
                id_to_mask = organize_files(mask_files, channel_prefix=channel_prefix, mask_suffix=mask_suffix)
                notifier( "Finished Organizing Files!")
                return id_to_image, id_to_mask
    return None

def copy_files_between_directories(source_dir, target_dir, file_types = None, event_manager: EventManager=None):
    file_filter = lambda file_path: file_path.is_file() and (True if file_types is None else file_path.suffix in file_types)


    files = listdir(source_dir)
    files_to_copy = [file for file in files if file_filter(file)]

    total_files = len(files_to_copy)
    copied_files = 0

    if event_manager is not None:
        event_manager.notify(
            event=ProgressEvent(0, process=f"Copy Files: {copied_files}/{total_files}"))
    for src_path in files_to_copy:
        target_path = target_dir / src_path.name

        try:
            if target_path.exists():
                if platform.system() == "Windows":
                    os.chmod(target_path, stat.S_IWRITE)
                else:
                    target_path.chmod(0o777)
                target_path.unlink()

            shutil.copy(str(src_path), str(target_path))

        except Exception as e:
            print(f"Something went wrong while processing {src_path.name}: {str(e)}")
        finally:
            copied_files+=1
            if event_manager is not None:
                event_manager.notify(event=ProgressEvent(int(copied_files/total_files*100), process=f"Copy Files: {copied_files}/{total_files}"))

    if event_manager is not None:
        event_manager.notify(
            event=ProgressEvent(100, process="Finished copy Files!"))

def load_lif3d_bioimage(lif3d_path):
    lif3d_path = pathlib.Path(lif3d_path)
    # See Readme at https://github.com/bioio-devs/bioio
    bio_img = BioImage(lif3d_path)

    # test_data = np.squeeze(bio_img.get_stack(), axis=1)
    images = []
    series_ids = []
    for scene in bio_img.scenes:
        bio_img.set_scene(scene)
        cur_data = bio_img.data
        shape = cur_data.shape

        if shape[-1] != 1024 or shape[-2] != 1024:
            # Special case of a single series in a .lif file having a resolution of 512 x 512
            continue

        cur_data = cur_data.reshape((shape[0], -1, shape[-2], shape[-1])).reshape(shape[0], shape[2], shape[1],
                                                                                  shape[-2], shape[-1])
        shape = cur_data.shape
        series_ids.append(scene)
        images.append(cur_data)
        # fig, axes = plt.subplots(ncols=shape[2], nrows=shape[1], sharex=True, sharey=True)
        # for iR in range(shape[1]):
        #    for iC in range(shape[2]):
        #        axes[iR, iC].imshow(test_data[0, iR, iC])
        # plt.tight_layout()
        # plt.show()
    images = np.stack(images)
    images = np.squeeze(images)
    return series_ids, images


def extract_from_lif3d_file(lif3d_path, target_dir, channel_prefix, event_manager: EventManager = None):
    series_ids, images = load_lif3d_bioimage(lif3d_path)

    images = np.transpose(images, axes=(0, 2, 3, 4, 1))
    target_dir = pathlib.Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)

    total_scenes = len(series_ids)
    if event_manager is not None:
        event_manager.notify(
            event=ProgressEvent(0, process=f"Extracting Series: {0}/{total_scenes}"))
    for s_idx, series in enumerate(images):
        series_id = series_ids[s_idx]
        for c_idx, channel_3d in enumerate(series):
            file_name = f"{series_id}{channel_prefix}{c_idx + 1}.tif"
            target_path = target_dir / file_name
            tifffile.imwrite(target_path, channel_3d)
        if event_manager is not None:
            event_manager.notify(event=ProgressEvent(int((s_idx + 1) / total_scenes * 100),
                                                     process=f"Extracted Series: {s_idx + 1}/{total_scenes}"))
    if event_manager is not None:
        event_manager.notify(
            event=ProgressEvent(100, process=f"Finished extracting Series!"))


def extract_from_lif_file(lif_path, target_dir,channel_prefix,event_manager: EventManager=None):
    """
    Extracts all series from the lif file using the bioio-lif library and
    copies the images to the target directory.
    Arguments:
          lif_path {str} -- The path to the lif file.
          target_dir {str} -- The path to the target directory.
    """

    lif_path = pathlib.Path(lif_path)
    target_dir = pathlib.Path(target_dir)
    if lif_path.suffix == ".lif":
        bio_image = BioImage(lif_path,reader=bioio_lif.Reader)  # Specify the backend explicitly
        data = np.squeeze(bio_image.data)
        is_3d = (data.ndim >= 4 and data.shape[1] > 1)

        if is_3d:
            extract_from_lif3d_file(lif_path, target_dir, channel_prefix, event_manager)
            return

        # Create the target directory if it doesn't exist
        target_dir.mkdir(parents=True, exist_ok=True)

        # get all series in the lif file
        scenes= bio_image.scenes
        total_scenes = len(scenes)
        if event_manager is not None:
            event_manager.notify(
                event=ProgressEvent(0, process=f"Extracting Series: {0}/{total_scenes}"))

        for index,scene_id in enumerate(scenes):
            scene= scene_id

            #remove the unnecessary data in the array
            bio_image.set_scene(scene)
            #TCZXY 5D array
            npy_array= bio_image.data
            squeezed_img= np.squeeze(npy_array)

            #get the amount of channels
            n_channels = squeezed_img.shape[0]

            for channel_id in range(n_channels):
                # Extract the height and width of the image
                image= squeezed_img[channel_id]
                img = Image.fromarray(image)#doesnt work

                # Construct file name and path
                file_name = f"{scene}{channel_prefix}{channel_id + 1}.tif"
                target_path = target_dir / file_name

                try:
                    # Handle existing files
                    if target_path.exists():
                        if platform.system() == "Windows":
                            os.chmod(target_path, stat.S_IWRITE)  # Set writable on Windows
                        else:
                            target_path.chmod(0o777)  # Set writable on Unix
                        target_path.unlink()  # Remove the existing file

                    # Save the image to the target path using pillows save function
                    img.save(target_path)

                except Exception as e:
                    print(f"Error processing {file_name}: {e}")
                    continue
            if event_manager is not None:
                event_manager.notify(event=ProgressEvent(int((index+1) / total_scenes * 100),
                                                         process=f"Extracted Series: {index+1}/{total_scenes}"))
        if event_manager is not None:
            event_manager.notify(
                event=ProgressEvent(100, process=f"Finished extracting Series!"))





def load_image_to_numpy(path):
    im = tifffile.imread(path)
    array = np.array(im)
    return array


def write_numpy_to_image(array, path):
    im = Image.fromarray(array)
    im.save(path)
    pass


def remove_gradient(img):

    """
    The method evens out the background of the images to prone microscopy errors

    Arguments:
        img {PIL.Image} -- The image to be corrected

    """
    top = np.median(img[100:200, 400: -400])
    bottom = np.median(img[-200:-100, 400: -400])

    left = np.median(img[400:-400, 100: 200])
    right = np.median(img[400:-400, -200: -100])

    median = np.median(img[200:-200, 200:-200])

    max_val = np.max([top, bottom, left, right])

    row_count = img.shape[0]

    X = np.arange(row_count) / (row_count - 1)
    b = bottom
    a = top - bottom
    Y_v = a * X + b
    Y_v -= median

    b = right
    a = left - right
    Y_h = a * X + b
    Y_h -= median

    correction_v = np.tile(Y_v, (row_count, 1)).transpose()
    correction_h = np.tile(Y_h, (row_count, 1))
    correction = correction_h + correction_v

    corrected_img = img + correction
    return corrected_img


def transform_image_path(image_path, output_path):
    """
    This method converts images with bit depth of 16 bit to 8 bit

    Attributes:
        image_path (pathlib.Path): Path to the image
        output_path (pathlib.Path): Path where to save the converted image

    Returns:
        True if the image was converted successfully
        False if the image was not converted because it had an incompatible format
    """
    with Image.open(image_path) as img:
        # convert to 8 bit if necessary
        if img.mode == "I;16":
            array16 = np.array(img, dtype=np.uint16)
            array8 = (array16 / 256).astype(np.uint8)
            img8 = Image.fromarray(array8)
            img8.save(output_path, format="TIFF")
            return True
        elif img.mode in ["L", "RGB", "P", "RGBA"]:
            return True
        else:
            return False


def process_channel(channel_id, channel_path):
    image = tifffile.imread(channel_path)
    if image.ndim == 3:
        image = np.max(image, axis=2)
    img = Image.fromarray(image)
    buffer = BytesIO()
    img.save(buffer, format="PNG")
    buffer.seek(0)

    return channel_id, base64.b64encode(buffer.getvalue()).decode('utf-8')

def convert_series_parallel(image_id, cur_image_paths):
    png_images = {image_id: {}}
    with ThreadPoolExecutor() as executor:
        futures = {
            executor.submit(process_channel, channel_id, cur_image_paths[channel_id]): channel_id
            for channel_id in cur_image_paths
        }
        for future in futures:
            channel_id, encoded_image = future.result()
            png_images[image_id][channel_id] = encoded_image

    return png_images

def convert_tiffs_to_png_parallel(image_paths):
    """
    Converts a dict of tiff images to png images using multiprocessing.

    Args:
        image_paths (dict): the dict of image paths of tiff images
    """
    if image_paths is not None:
        png_images = {}
        with ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(convert_series_parallel, image_id, image_paths[image_id]): image_id
                for image_id in image_paths
            }
            for future in futures:
                result = future.result()
                png_images.update(result)

        return png_images
    else:
        return None

def convert_tiffs_to_png(image_paths):
    """
    Converts a dict of tiff images to png images.

    Args:
        image_paths (dict): the dict of image paths of tiff images
    """
    if image_paths is not None:
        png_images = {}
        for image_id in image_paths:
            cur_image_paths = image_paths[image_id]
            if image_id not in png_images:
                png_images[image_id] = {}
            for channel_id in cur_image_paths:
                image = image = Image.open(cur_image_paths[channel_id])

                buffer = BytesIO()
                image.save(buffer, format="PNG")
                buffer.seek(0)

                png_images[image_id][channel_id] = base64.b64encode(buffer.getvalue()).decode('utf-8')

        return png_images
    else:
        return None


import os

import numpy as np
from cv2 import imread
from tifffile import tifffile
from PIL import Image

from cellsepi.backend.main_window.data_util import load_directory, ReturnTypePath
from cellsepi.backend.main_window.expert_mode.listener import ProgressEvent
from cellsepi.backend.main_window.expert_mode.module import *
from cellsepi.backend.main_window.images import BatchImageSegmentation
from cellsepi.frontend.main_window.gui_directory import DirectoryCard


class Project3dTo2d(Module, ABC):
    _gui_config = ModuleGuiConfig("Project3Dto2D",Categories.FILTERS,"This modules handles the conversion from 3D data to 2D data based on an max z projection.")
    def __init__(self, module_id: str) -> None:
        super().__init__(module_id)
        self.inputs = {
            "image_paths": InputPort("image_paths", dict),
        }
        self.outputs = {
            "image_paths": OutputPort("image_paths", dict),
        }

    def run(self):
        images = self.inputs["image_paths"].data
        outputs_images = {}
        n_series = len(images)
        self.event_manager.notify(ProgressEvent(percent=0, process=f"Projecting Series: Starting"))
        for iN,series in enumerate(images):
            outputs_images[series] = {}
            for channel in images[series]:
                image_path = images[series][channel]
                image = tifffile.imread(image_path) #dimensions are: X, Y, Z
                image_max = np.max(image,axis=2)
                base_dir = os.path.dirname(image_path)
                proj_dir = os.path.join(base_dir, "projections")
                os.makedirs(proj_dir, exist_ok=True)
                name =os.path.basename(image_path)
                new_path = os.path.join(proj_dir, name)
                image_max_8bit = ((image_max - image_max.min()) / (image_max.max() - image_max.min()) * 255).astype(np.uint8)
                img8 = Image.fromarray(image_max_8bit)
                img8.save(new_path, format="TIFF")
                outputs_images[series][channel] = new_path

            self.event_manager.notify(ProgressEvent(percent=int((iN+1) / n_series * 100), process=f"Projecting Series: {iN+1}/{n_series}"))
        self.outputs["image_paths"].data = outputs_images
        self.event_manager.notify(ProgressEvent(percent=100, process=f"Projecting Series: Finished"))







import multiprocessing
import sys

import flet as ft
from cellsepi.frontend.main_window.gui import GUI
from cellsepi.cli import build as flet_build

def main():
    if len(sys.argv) > 1 and sys.argv[1].lower() == "build":
        flet_build()
    else:
        multiprocessing.set_start_method("spawn")
        ft.app(target=async_main, view=ft.FLET_APP)

async def async_main(page: ft.Page):
    gui = GUI(page)
    gui.build()
